@@ -29,7 +29,7 @@ print OUT <<'!NO!SUBS!';
 # Info-URL: http://degraaff.org/checkbot/
 # Comments to: checkbot@degraaff.org
 #
-# $Id: checkbot.pl,v 1.64 2001/04/15 19:34:30 graaff Exp $
+# $Id: checkbot.pl,v 1.65 2001/10/21 08:07:19 graaff Exp $
 # (Log information can be found at the end of the script)
 
 require 5.004;
@@ -48,7 +48,7 @@ Checkbot - WWW Link Verifier
 checkbot [B<--debug>] [B<--help>] [B<--verbose>] [B<--url> start URL] 
          [B<--match> match string] [B<--exclude> exclude string]
          [B<--proxy> proxy URL] [B<--internal-only>]
-         [B<--ignore> ignore string] [B<-file> file name]
+         [B<--ignore> ignore string] [B<--file> file name]
          [B<--style> style file URL]
          [B<--mailto> email address]
          [B<--note> note] [B<--sleep> seconds] [B<--timeout> timeout]
@@ -172,7 +172,7 @@ layout of pages generated by Checkbot.
 
 =item --dontwarn <HTTP response codes regular expression>
 
-Don't include warnings on the result pages for those HTTP response
+Do not include warnings on the result pages for those HTTP response
 codes which match the regular expression. For instance, --dontwarn
 "(301|404)" would not include 301 and 404 response codes.
 
@@ -222,7 +222,7 @@ my %problems = ();
 
 # Version information
 my $VERSION;
-( $VERSION ) = sprintf("%d.%02d", q$Revision: 1.64 $ =~ /(\d+)\.(\d+)/);
+( $VERSION ) = sprintf("%d.%02d", q$Revision: 1.65 $ =~ /(\d+)\.(\d+)/);
 
 # If on a Mac we should ask for the arguments through some MacPerl stuff
 if ($^O eq 'MacOS') {
@@ -285,6 +285,19 @@ sub check_options {
 
 sub init_modules {
 
+# Using keep-alive from HTTP 1.1 will be a great efficiency
+# improvement, but it wont be available until libwww-perl 5.54. This
+# code hints at what should be done, but it's not an actual
+# implementation.
+
+#      if ( $server->{keep_alive} ) {
+#           eval {
+#               require LWP::Protocol::http11;
+#               LWP::Protocol::implementor('http', 'LWP::Protocol::http11');
+#           };
+#           warn "Cannot use keep alives, please upgrade LWP -- $@\n" if $@;
+#      }
+
   use URI;
   # Prepare the user agent to be used:
   use LWP::UserAgent;
@@ -295,10 +308,6 @@ sub init_modules {
   $main::ua->agent("Checkbot/$VERSION LWP/" . LWP::Version);
   $main::ua->timeout($main::opt_timeout);
 
-  # Don't use strict parsing, it is very likely we will
-  # encounter invalid HTML documents on the way.
-  URI::URL::strict(0);
-
   require Mail::Send if defined $main::opt_mailto;
 
   use HTTP::Status;
@@ -334,7 +343,7 @@ sub init_globals {
   # just process URLs in order as they appear on the command line.
   unshift(@ARGV, $main::opt_url) if $main::opt_url;
   foreach (@ARGV) {
-    $url = new URI::URL $_;
+    $url = URI->new($_);
     # If no scheme is defined we will assume file is used, so that
     # it becomes easy to check a single file.
     $url->scheme('file') unless defined $url->scheme;
@@ -369,8 +378,7 @@ sub init_globals {
 
   # We write out our status every now and then.
   $main::cp_int = 1;
-  $main::cp_last = time();
-
+  $main::cp_last = 0;
 }
 
 
@@ -465,7 +473,7 @@ sub handle_url {
   # if it is a duplicate
   return if &add_checked($urlstr);
 
-  my $url = new URI::URL $urlstr;
+  my $url = URI->new($urlstr);
   $main::st_int[$main::LINKS]++;
 	
   if (defined($url->scheme) 
@@ -600,92 +608,12 @@ sub check_external {
 
     next if $prevurl eq $newurl;
     $prevurl = $newurl;
-    
-    my $url = new URI::URL $newurl, $urlparent;
+
+    my $url = URI->new_abs($newurl, $urlparent);
 
     if ($url->scheme =~ /^(http|file|ftp|gopher|nntp)$/o) { 
 
-      # Normally, we would only need to do a HEAD, but given the way
-      # LWP handles gopher requests, we need to do a GET on those to
-      # get at least a 500 and 501 error. We would need to parse the
-      # document returned by LWP to find out if we had problems
-      # finding the file. -- Patch by Bruce Speyer
-      # <bspeyer@texas-one.org>
-
-      # We also need to do GET instead of HEAD if we know the remote server
-      # won't accept it.  The standard way for an HTTP server to indicate
-      # this is by returning a 405 ("Method Not Allowed") or 501 ("Not
-      # Implemented").  Other circumstances may also require sending GETs
-      # instead of HEADs to a server.  Details are documented below.
-      # -- Larry Gilbert <larry@n2h2.com>
-
-      my ($reqtype, $ref_header, $request, $response);
-
-      foreach $reqtype ('HEAD', 'GET') {
-        # Under normal circumstances, we will go through this loop only once
-        # and thus make a single HEAD request.
-
-        # If we already know a HEAD won't work, skip it and do GET
-        next if $reqtype eq 'HEAD' && ($url->scheme =~ /^gopher$/o ||
-         $main::servers_get_only{$url->netloc} );
-
-        print STDERR "     $reqtype ", $url->abs, "\n" if $main::opt_verbose;
-        $ref_header = new HTTP::Headers 'Referer' => $urlparent;
-        $request = new HTTP::Request($reqtype, $url, $ref_header);
-        $response = $main::ua->simple_request($request);
-
-        if ($reqtype eq 'HEAD') {
-
-          # 405 and 501 are standard indications that HEAD shouldn't be used
-          if ($response->code =~ /^(405|501)$/o) {
-            print STDERR "Server doesn't like HEAD requests; retrying\n"
-             if $main::opt_verbose;
-            $main::servers_get_only{$url->netloc}++;
-            next;
-          }
-
-          # Microsoft IIS has been seen dropping the connection prematurely
-          # when it should be returning 405 instead
-          elsif ($response->status_line =~ /^500 unexpected EOF/o) {
-            print STDERR "Server hung up on HEAD request; retrying\n"
-             if $main::opt_verbose;
-            $main::servers_get_only{$url->netloc}++;
-            next;
-          }
-
-          # Netscape Enterprise has been seen returning 500 and even 404
-          # (yes, 404!!) in response to HEAD requests
-          elsif (defined $response->server and $response->server =~ /^Netscape-Enterprise/o &&
-           $response->code =~ /^(404|500)$/o) {
-            print STDERR "Unreliable response to HEAD request; retrying\n"
-             if $main::opt_verbose;
-            $main::servers_get_only{$url->netloc}++;
-            next;
-          }
-
-          # JavaWebServer/1.1.3 has been seen returning 500
-          elsif (defined $response->server and $response->server =~ /^JavaWebServer/o &&
-           $response->code =~ /^500$/o) {
-            print STDERR "Unreliable response to HEAD request; retrying\n"
-             if $main::opt_verbose;
-            $main::servers_get_only{$url->netloc}++;
-            next;
-          }
-
-          # ALEPH has been seen returning 500
-          elsif (defined $response->server and $response->server =~ /^ALEPH/o &&
-           $response->code =~ /^500$/o) {
-            print STDERR "Unreliable response to HEAD request; retrying\n"
-             if $main::opt_verbose;
-            $main::servers_get_only{$url->netloc}++;
-            next;
-          }
-
-          # If a HEAD request resulted in nothing noteworthy, no need for
-          # any further attempts
-          else { last; }
-        }
-      }
+      my $response = get_headers($url, $urlparent);
 
       if ($response->is_error || $response->is_redirect) {
 	if (defined $main::opt_ignore && $url =~ /$main::opt_ignore/o) {
@@ -839,7 +767,7 @@ sub get_server_type {
       my $response = $main::ua->simple_request($request);
       $result = $response->header('Server');
     }
-    $result = "Unknown server type" if $result eq "";
+    $result = "Unknown server type" if ! defined $result or $result eq "";
     print STDERR "=== Server $server is a $result\n" if $main::opt_verbose;
     $main::server_type{$server} = $result;
   }
@@ -928,6 +856,8 @@ sub handle_doc {
 	if ($url =~ /^(http|ftp):/o) {
 	  print EXTERNAL $url . "|" . $response->base . "\n";
 	  $doc_ext++;
+	} else {
+	  print STDERR "--> Ignored: external URL is not HTTP or FTP: $url\n" if $main::opt_verbose;
 	}
       }
     }
@@ -944,6 +874,101 @@ sub handle_doc {
   }
 }
 
+sub get_headers {
+
+# A better solution here would be to use GET exclusively. Here is how
+# to do that. We would have to set this max_size thing in
+# check_external, I guess...
+
+# Set $ua->max_size(1) and then try a normal GET request.
+
+  my ($url, $urlparent) = @_;
+
+  # Normally, we would only need to do a HEAD, but given the way LWP
+  # handles gopher requests, we need to do a GET on those to get at
+  # least a 500 and 501 error. We would need to parse the document
+  # returned by LWP to find out if we had problems finding the
+  # file. -- Patch by Bruce Speyer <bspeyer@texas-one.org>
+
+  # We also need to do GET instead of HEAD if we know the remote
+  # server won't accept it.  The standard way for an HTTP server to
+  # indicate this is by returning a 405 ("Method Not Allowed") or 501
+  # ("Not Implemented").  Other circumstances may also require sending
+  # GETs instead of HEADs to a server.  Details are documented below.
+  # -- Larry Gilbert <larry@n2h2.com>
+
+  my ($reqtype, $ref_header, $request, $response);
+
+  foreach $reqtype ('HEAD', 'GET') {
+    # Under normal circumstances, we will go through this loop only once
+    # and thus make a single HEAD request.
+
+    # If we already know a HEAD won't work, skip it and do GET
+    next if $reqtype eq 'HEAD' && ($url->scheme =~ /^gopher$/o ||
+				   $main::servers_get_only{$url->netloc} );
+
+    print STDERR "     $reqtype ", $url->abs, "\n" if $main::opt_verbose;
+    $ref_header = new HTTP::Headers 'Referer' => $urlparent;
+    $request = new HTTP::Request($reqtype, $url, $ref_header);
+    $response = $main::ua->simple_request($request);
+
+    if ($reqtype eq 'HEAD') {
+
+      # 405 and 501 are standard indications that HEAD shouldn't be used
+      if ($response->code =~ /^(405|501)$/o) {
+	print STDERR "Server doesn't like HEAD requests; retrying\n"
+	  if $main::opt_verbose;
+	$main::servers_get_only{$url->netloc}++;
+	next;
+      }
+
+      # Microsoft IIS has been seen dropping the connection prematurely
+      # when it should be returning 405 instead
+      elsif ($response->status_line =~ /^500 unexpected EOF/o) {
+	print STDERR "Server hung up on HEAD request; retrying\n"
+	  if $main::opt_verbose;
+	$main::servers_get_only{$url->netloc}++;
+	next;
+      }
+
+      # Netscape Enterprise has been seen returning 500 and even 404
+      # (yes, 404!!) in response to HEAD requests
+      elsif (defined $response->server and $response->server =~ /^Netscape-Enterprise/o &&
+	     $response->code =~ /^(404|500)$/o) {
+	print STDERR "Unreliable response to HEAD request; retrying\n"
+	  if $main::opt_verbose;
+	$main::servers_get_only{$url->netloc}++;
+	next;
+      }
+
+      # JavaWebServer/1.1.3 has been seen returning 500
+      elsif (defined $response->server and $response->server =~ /^JavaWebServer/o &&
+	     $response->code =~ /^500$/o) {
+	print STDERR "Unreliable response to HEAD request; retrying\n"
+	  if $main::opt_verbose;
+	$main::servers_get_only{$url->netloc}++;
+	next;
+      }
+
+      # ALEPH has been seen returning 500
+      elsif (defined $response->server and $response->server =~ /^ALEPH/o &&
+	     $response->code =~ /^500$/o) {
+	print STDERR "Unreliable response to HEAD request; retrying\n"
+	  if $main::opt_verbose;
+	$main::servers_get_only{$url->netloc}++;
+	next;
+      }
+
+      # If a HEAD request resulted in nothing noteworthy, no need for
+      # any further attempts
+      else { last; }
+    }
+  }
+
+  return $response;
+}
+
+
 sub add_to_queue {
   my ($url, $parent) = @_;
 
@@ -1103,15 +1128,15 @@ sub check_point {
 
 sub send_mail {
   my $msg = new Mail::Send;
-  my $sub = 'Checkbot results are available';
-  $sub .= " for $main::opt_url" if defined $main::opt_url;
+  my $sub = 'Checkbot results for ';
+  $sub .= join(', ', @starturls);
 
   $msg->to($main::opt_mailto);
   $msg->subject($sub);
 
   my $fh = $msg->open;
 
-  print $fh "$sub.\n\n";
+  print $fh "Checkbot results for:\n  " . join("\n  ", @starturls) . "\n\n";
   print $fh "User-supplied note: $main::opt_note\n\n"
     if defined $main::opt_note;
 
@@ -1187,7 +1212,7 @@ sub count_problems {
   foreach my $code (sort keys %problems) {
     foreach my $url (sort keys %{ $problems{$code}}) {
       foreach my $response (@{ $problems{$code}{$url} }) {
-	my $parent = new URI::URL $response->request->header('Referer');
+	my $parent = URI->new($response->request->header('Referer'));
 	$count++ if $parent =~ $server;
       }
     }
@@ -1198,6 +1223,11 @@ sub count_problems {
 
 
 # $Log: checkbot.pl,v $
+# Revision 1.65  2001/10/21 08:07:19  graaff
+#  * Move to URI from URI::URL
+#  * Start writing results pages right at the start
+#  * A few small bugs fixed (see ChangeLog)
+#
 # Revision 1.64  2001/04/15 19:34:30  graaff
 # Fix printing of starting URLs in email, removed duplicate header in
 # report, try more environment variables to set temporary directory,
